#include <stdio.h>
/* malloc, rand */
#include <stdlib.h>
/* memcpy */
#include <string.h>

#include "rule.h"

#include "monster.h"

#define WORST_SCORE (-9999)

monster *
monster_new(int layer_size)
{
  int i;

  monster *p = NULL;

  if (layer_size <= 0)
  {
    fprintf(stderr, "monster_new: layer_size is non-positive (%d)\n",
            layer_size);
    return NULL;
  }

  p = (monster *) malloc(sizeof(monster));
  if (p == NULL)
  {
    fprintf(stderr, "monster_new: malloc(p) failed\n");
    return NULL;
  }

  p->layer_size = layer_size;
  p->state = NULL;
  p->score = NULL;
  p->layer_top = -1;
  p->turn_player = NULL;
  p->grid_from = NULL;
  p->grid_to = NULL;

  p->state = (int **) malloc(sizeof(int *) * p->layer_size);
  if (p->state == NULL)
  {
    fprintf(stderr, "monster_new: malloc(p->state) failed\n");
    monster_delete(p);
    p = NULL;
    return NULL;
  }
  for (i = 0; i < p->layer_size; i++)
    p->state[i] = NULL;

  for (i = 0; i < p->layer_size; i++)
  {
    p->state[i] = (int *) malloc(sizeof(int) * STATE_SIZE);
    if (p->state[i] == NULL)
    {
      fprintf(stderr, "monster_new: malloc(p->state[%d]) failed\n", i);
      monster_delete(p);
      p = NULL;
      return NULL;
    }
  }

  p->score = (int *) malloc(sizeof(int) * p->layer_size);
  if (p->score == NULL)
  {
    fprintf(stderr, "monster_new: malloc(p->score) failed\n");
    monster_delete(p);
    p = NULL;
    return NULL;
  }

  p->turn_player = (int *) malloc(sizeof(int) * p->layer_size);
  if (p->turn_player == NULL)
  {
    fprintf(stderr, "monster_new: malloc(p->turn_player) failed\n");
    monster_delete(p);
    p = NULL;
    return NULL;
  }

  p->grid_from = (int **) malloc(sizeof(int *) * p->layer_size);
  if (p->grid_from == NULL)
  {
    fprintf(stderr, "monster_new: malloc(p->grid_from) failed\n");
    monster_delete(p);
    p = NULL;
    return NULL;
  }
  for (i = 0; i < p->layer_size; i++)
    p->grid_from[i] = NULL;

  for (i = 1; i < p->layer_size; i++)
  {
    p->grid_from[i] = (int *) malloc(sizeof(int) * (i + 1));
    if (p->grid_from[i] == NULL)
    {
      fprintf(stderr, "monster_new: malloc(p->grid_from[%d]) failed\n", i);
      monster_delete(p);
      p = NULL;
      return NULL;
    }
  }

  p->grid_to = (int **) malloc(sizeof(int *) * p->layer_size);
  if (p->grid_to == NULL)
  {
    fprintf(stderr, "monster_new: malloc(p->grid_to) failed\n");
    monster_delete(p);
    p = NULL;
    return NULL;
  }
  for (i = 0; i < p->layer_size; i++)
    p->grid_to[i] = NULL;

  for (i = 1; i < p->layer_size; i++)
  {
    p->grid_to[i] = (int *) malloc(sizeof(int) * (i + 1));
    if (p->grid_to[i] == NULL)
    {
      fprintf(stderr, "monster_new: malloc(p->grid_to[%d]) failed\n", i);
      monster_delete(p);
      p = NULL;
      return NULL;
    }
  }

  return p;
}

void
monster_delete(monster *p)
{
  int i;

  if (p == NULL)
    return;

  if (p->state != NULL)
  {
    for (i = 0; i < p->layer_size; i++)
    {
      if (p->state[i] != NULL)
      {
        free(p->state[i]);
        p->state[i] = NULL;
      }
    }

    free(p->state);
    p->state = NULL;
  }

  if (p->score != NULL)
  {
    free(p->score);
    p->score = NULL;
  }

  if (p->turn_player != NULL)
  {
    free(p->turn_player);
    p->turn_player = NULL;
  }

  if (p->grid_from != NULL)
  {
    for (i = 1; i < p->layer_size; i++)
    {
      if (p->grid_from[i] != NULL)
      {
        free(p->grid_from[i]);
        p->grid_from[i] = NULL;
      }
    }

    free(p->grid_from);
    p->grid_from = NULL;
  }

  if (p->grid_to != NULL)
  {
    for (i = 1; i < p->layer_size; i++)
    {
      if (p->grid_to[i] != NULL)
      {
        free(p->grid_to[i]);
        p->grid_to[i] = NULL;
      }
    }

    free(p->grid_to);
    p->grid_to = NULL;
  }

  free(p);
  p = NULL;
}
