#include <stdio.h>
/* malloc, rand */
#include <stdlib.h>
/* memcpy */
#include <string.h>

#include "world.h"
#include "rule.h"
#include "monster.h"

#include "think.h"

#define WORST_SCORE (-9999)

static int state_score(int *state, int turn_player, int victory_condition);

int
monster_init(world *wp, int layer_top)
{
  int i;

  if (wp == NULL)
  {
    fprintf(stderr, "monster_init: wp is NULL\n");
    return 1;
  }
  if (layer_top >= wp->mp->layer_size)
  {
    fprintf(stderr, "monster_init: layer_top is too big "
            "(layer_top = %d, wp->mp->layer_size = %d)\n",
            layer_top, wp->mp->layer_size);
    return 1;
  }
  if (layer_top < 1)
  {
    fprintf(stderr, "monster_init: layer_top must be 1 or greater "
            "(layer_top = %d)",
            layer_top);
    return 1;
  }

  wp->mp->layer_top = layer_top;
  memcpy(wp->mp->state[wp->mp->layer_top], wp->state,
         sizeof(int) * STATE_SIZE);
  wp->mp->victory_condition = wp->victory_condition;

  for (i = 0; i <= wp->mp->layer_top; i++)
  {
    wp->mp->turn_player[wp->mp->layer_top - i] = (wp->turn_player + i) % 2;
  }

  for (i = 1; i <= wp->mp->layer_top; i++)
  {    
    wp->mp->grid_from[wp->mp->layer_top][i] = -1;
    wp->mp->grid_to[wp->mp->layer_top][i] = -1;
  }

  return 0;
}

static int
state_score(int *state, int turn_player, int victory_condition)
{
  int num_friendly;
  int num_enemy;

  if (turn_player == 0)
  {
    num_friendly = state[99];
    num_enemy = state[100];
  }
  else
  {
    num_friendly = state[100];
    num_enemy = state[99];
  }

  if (num_friendly < victory_condition)
    return WORST_SCORE;
  else if (num_enemy < victory_condition)
    return -WORST_SCORE;

  return num_friendly - num_enemy;
}

int
monster_think(monster *p, int layer_now)
{
  int i;
  int j;
  int g;
  int x;
  int y;
  int g0;
  int x0;
  int y0;
  int dx;
  int dy;
  int turn_player;
  int head_friendly;
  int num_friendly;
  int num_enemy;
  int score;
  int score_next;
  int score_next_temp;
  int count_same_score;
  int pick_this;
  int strictly_better;
  int not_bad;
  int *state = NULL;

  state = p->state[layer_now];

  turn_player = p->turn_player[layer_now];
  if (turn_player == 0)
  {
    head_friendly = 0;
    num_friendly = state[99];
    num_enemy = state[100];
  }
  else
  {
    head_friendly = 9;
    num_friendly = state[100];
    num_enemy = state[99];
  }

  if (num_friendly < p->victory_condition)
  {
    p->score[layer_now] = WORST_SCORE;
    p->grid_from[layer_now][layer_now] = -1;
    p->grid_to[layer_now][layer_now] = -1;
    return 0;
  }
  if (num_enemy < p->victory_condition)
  {
    p->score[layer_now] = -WORST_SCORE;
    p->grid_from[layer_now][layer_now] = -1;
    p->grid_to[layer_now][layer_now] = -1;
    return 0;
  }

  /* don't use INT_MIN here because -INT_MIN > INT_MAX*/
  p->score[layer_now] = WORST_SCORE;
  count_same_score = 0;
  score_next = WORST_SCORE;
  score_next_temp = WORST_SCORE;

  for (i = 0; i < 9; i++)
  {
    if (state[head_friendly + i] < 0)
      continue;

    g0 = state[head_friendly + i];
    x0 = g0 % 9;
    y0 = g0 / 9;

    for (j = 0; j < 4; j++)
    {
      switch (j)
      {
      case 0:
        dx = 1;
        dy = 0;
        break;
      case 1:
        dx = 0;
        dy = 1;
        break;
      case 2:
        dx = -1;
        dy = 0;
        break;
      default:
        dx = 0;
        dy = -1;
        break;
      }

      x = x0 + dx;
      y = y0 + dy;

      while ((x >= 0) && (x < 9) && (y >= 0) && (y < 9))
      {
        g = y * 9 + x;

        if (state[18 + g] >= 0)
          break;

        memcpy(p->state[layer_now - 1], p->state[layer_now],
               sizeof(int) * STATE_SIZE);
        move_piece(p->state[layer_now - 1], g0, g);
        capture_piece(p->state[layer_now - 1], g);

        if (layer_now >= 2)
        {
          monster_think(p, layer_now - 1);
        }
        else
        {
          p->score[0] = state_score(p->state[0], p->turn_player[0],
                                    p->victory_condition);
        }
        score = -(p->score[layer_now - 1]);
        if (layer_now < p->layer_top)
        {
          if (-score < p->score[layer_now + 1])
          {
            p->score[layer_now] = score;
            return 0;
          }
          if ((layer_now < p->layer_top - 1)
              && (-score == p->score[layer_now + 1]))
          {
            p->score[layer_now] = score;

            p->grid_from[layer_now][layer_now] = g0;
            p->grid_to[layer_now][layer_now] = g;
            if (layer_now >= 2)
            {
              memcpy(p->grid_from[layer_now], p->grid_from[layer_now - 1],
                     sizeof(int) * layer_now);
              memcpy(p->grid_to[layer_now], p->grid_to[layer_now - 1],
                     sizeof(int) * layer_now);
            }

            return 0;
          }
        }
        if (score >= p->score[layer_now])
        {
          pick_this = 0;
          strictly_better = 0;
          not_bad = 1;

          if (layer_now >= 2)
          {
            score_next_temp = -state_score(p->state[layer_now - 1],
                                           p->turn_player[layer_now - 1],
                                           p->victory_condition);
          }

          if (score > p->score[layer_now])
          {
            strictly_better = 1;
          }
          else
          {
            if (layer_now >= 2)
            {
              if (score_next_temp > score_next)
              {
                strictly_better = 1;
              }
              else if (score_next_temp < score_next)
              {
                not_bad = 0;
              }
            }
          }

          if (strictly_better)
          {
            count_same_score = 1;
            pick_this = 1;
          }
          else if ((layer_now == p->layer_top) && (not_bad))
          {
            count_same_score++;
            if ((count_same_score <= 1)
                || (rand() % count_same_score == 0))
            { 
              pick_this = 1;
            }
          }

          if (pick_this)
          {
            p->score[layer_now] = score;
            p->grid_from[layer_now][layer_now] = g0;
            p->grid_to[layer_now][layer_now] = g;
            if (layer_now >= 2)
            {
              score_next = score_next_temp;

              memcpy(p->grid_from[layer_now], p->grid_from[layer_now - 1],
                     sizeof(int) * layer_now);
              memcpy(p->grid_to[layer_now], p->grid_to[layer_now - 1],
                     sizeof(int) * layer_now);
            }
          }
        }

        x += dx;
        y += dy;
      }
    }
  }

  return 0;
}
