#include <stdio.h>
/* malloc */
#include <stdlib.h>

#include "monster.h"
#include "mode.h"
#include "rule.h"

#include <curses.h>

#include "world.h"

world *
world_new(void)
{
  world *wp = NULL;

  wp = (world *) malloc(sizeof(world));
  if (wp == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp) failed\n");
    return NULL;
  }

  wp->state = NULL;
  wp->turn_player = 0;
  wp->grid_from = -1;
  wp->grid_to = -1;
  wp->cursor_x = 4;
  wp->cursor_y = 4;
  wp->last_move = -1;
  wp->player = NULL;
  wp->victory_condition = 5;
  wp->mp = NULL;
  wp->scr = NULL;
  wp->scr_size_x = 0;
  wp->scr_size_y = 0;
  wp->should_quit = 0;
  wp->mode = MODE_BEFORE_GAME;
  wp->option_item = OPTION_ITEM_BLUE;
  wp->waiting_for_move = 1;
  wp->cheat_peek_monster = 0;

  wp->state = (int *) malloc(sizeof(int) * STATE_SIZE);
  if (wp->state == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->state) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }

  wp->player = (int *) malloc(sizeof(int) * 2);
  if (wp->player == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->player) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  wp->player[0] = 0;
  wp->player[1] = 4;

  wp->mp = monster_new(5 + 1);
  if (wp->mp == NULL)
  {
    fprintf(stderr, "world_new: monster_new failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }

  return wp;
}

void
world_delete(world *wp)
{
  if (wp == NULL)
    return;

  if (wp->state != NULL)
  {
    free(wp->state);
    wp->state = NULL;
  }

  if (wp->player != NULL)
  {
    free(wp->player);
    wp->player = NULL;
  }

  if (wp->mp != NULL)
  {
    monster_delete(wp->mp);
    wp->mp = NULL;
  }

  free(wp);
  wp = NULL;
}

int
world_reset(world *wp)
{
  if (wp == NULL)
  {
    fprintf(stderr, "world_reset: wp is NULL\n");
    return 1;
  }

  state_reset(wp->state);
  wp->turn_player = 0;
  wp->grid_from = -1;
  wp->grid_to = -1;
  wp->cursor_x = 4;
  wp->cursor_y = 4;
  wp->last_move = -1;
  wp->mp->layer_top = -1;
  wp->mode = MODE_BEFORE_GAME;
  wp->waiting_for_move = 1;

  return 0;
}

